
# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "func"

@_ods_cext.register_operation(_Dialect)
class CallIndirectOp(_ods_ir.OpView):
  r"""
  The `func.call_indirect` operation represents an indirect call to a value
  of function type. The operands and result types of the call must match the
  specified function type.
  
  Function values can be created with the
  [`func.constant` operation](#funcconstant-constantop).
  
  Example:
  
  ```mlir
  %func = func.constant @my_func : (tensor<16xf32>, tensor<16xf32>) -> tensor<16xf32>
  %result = func.call_indirect %func(%0, %1) : (tensor<16xf32>, tensor<16xf32>) -> tensor<16xf32>
  ```
  """

  OPERATION_NAME = "func.call_indirect"

  _ODS_REGIONS = (0, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], callee: _ods_ir.Value, callee_operands: _Sequence[_ods_ir.Value], *, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(callee)
    operands.extend(_get_op_results_or_values(callee_operands))
    _ods_context = _ods_get_default_loc_context(loc)
    if arg_attrs is not None: attributes["arg_attrs"] = (arg_attrs if (
        isinstance(arg_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(arg_attrs, context=_ods_context))
    if res_attrs is not None: attributes["res_attrs"] = (res_attrs if (
        isinstance(res_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(res_attrs, context=_ods_context))
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def callee(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def callee_operands(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
    return self.operation.operands[1:1 + _ods_variadic_group_length]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["arg_attrs"]

  @arg_attrs.setter
  def arg_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["arg_attrs"] = value
    elif "arg_attrs" in self.operation.attributes:
      del self.operation.attributes["arg_attrs"]

  @arg_attrs.deleter
  def arg_attrs(self):
    del self.operation.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["res_attrs"]

  @res_attrs.setter
  def res_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["res_attrs"] = value
    elif "res_attrs" in self.operation.attributes:
      del self.operation.attributes["res_attrs"]

  @res_attrs.deleter
  def res_attrs(self):
    del self.operation.attributes["res_attrs"]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(CallIndirectOp)
class CallIndirectOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "func.call_indirect"

  @builtins.property
  def callee(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def callee_operands(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 2 + 1
    return self.operands[1:1 + _ods_variadic_group_length]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.attributes:
      return None
    return self.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.attributes:
      return None
    return self.attributes["res_attrs"]

def call_indirect(results_: _Sequence[_ods_ir.Type], callee: _ods_ir.Value, callee_operands: _Sequence[_ods_ir.Value], *, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, CallIndirectOp]:
  op = CallIndirectOp(results_=results_, callee=callee, callee_operands=callee_operands, arg_attrs=arg_attrs, res_attrs=res_attrs, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class CallOp(_ods_ir.OpView):
  r"""
  The `func.call` operation represents a direct call to a function that is
  within the same symbol scope as the call. The operands and result types of
  the call must match the specified function type. The callee is encoded as a
  symbol reference attribute named "callee".
  
  Example:
  
  ```mlir
  %2 = func.call @my_add(%0, %1) : (f32, f32) -> f32
  ```
  """

  OPERATION_NAME = "func.call"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _Sequence[_ods_ir.Type], callee: _Union[str, _ods_ir.FlatSymbolRefAttr], operands_: _Sequence[_ods_ir.Value], *, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, no_inline: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(operands_))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["callee"] = (callee if (
    isinstance(callee, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('FlatSymbolRefAttr')) else
      _ods_ir.AttrBuilder.get('FlatSymbolRefAttr')(callee, context=_ods_context))
    if arg_attrs is not None: attributes["arg_attrs"] = (arg_attrs if (
        isinstance(arg_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(arg_attrs, context=_ods_context))
    if res_attrs is not None: attributes["res_attrs"] = (res_attrs if (
        isinstance(res_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(res_attrs, context=_ods_context))
    if bool(no_inline): attributes["no_inline"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.extend(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def callee(self) -> _ods_ir.FlatSymbolRefAttr:
    return self.operation.attributes["callee"]

  @callee.setter
  def callee(self, value: _ods_ir.FlatSymbolRefAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["callee"] = value

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["arg_attrs"]

  @arg_attrs.setter
  def arg_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["arg_attrs"] = value
    elif "arg_attrs" in self.operation.attributes:
      del self.operation.attributes["arg_attrs"]

  @arg_attrs.deleter
  def arg_attrs(self):
    del self.operation.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["res_attrs"]

  @res_attrs.setter
  def res_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["res_attrs"] = value
    elif "res_attrs" in self.operation.attributes:
      del self.operation.attributes["res_attrs"]

  @res_attrs.deleter
  def res_attrs(self):
    del self.operation.attributes["res_attrs"]

  @builtins.property
  def no_inline(self) -> bool:
    return "no_inline" in self.operation.attributes

  @no_inline.setter
  def no_inline(self, value):
    if bool(value):
      self.operation.attributes["no_inline"] = _ods_ir.UnitAttr.get()
    elif "no_inline" in self.operation.attributes:
      del self.operation.attributes["no_inline"]

  @no_inline.deleter
  def no_inline(self):
    del self.operation.attributes["no_inline"]

@_ods_cext.register_op_adaptor(CallOp)
class CallOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "func.call"

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def callee(self) -> _ods_ir.FlatSymbolRefAttr:
    return self.attributes["callee"]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.attributes:
      return None
    return self.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.attributes:
      return None
    return self.attributes["res_attrs"]

  @builtins.property
  def no_inline(self) -> bool:
    return "no_inline" in self.attributes

def call(result: _Sequence[_ods_ir.Type], callee: _Union[str, _ods_ir.FlatSymbolRefAttr], operands_: _Sequence[_ods_ir.Value], *, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, no_inline: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, CallOp]:
  op = CallOp(result=result, callee=callee, operands_=operands_, arg_attrs=arg_attrs, res_attrs=res_attrs, no_inline=no_inline, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ConstantOp(_ods_ir.OpView):
  r"""
  The `func.constant` operation produces an SSA value from a symbol reference
  to a `func.func` operation
  
  Example:
  
  ```mlir
  // Reference to function @myfn.
  %2 = func.constant @myfn : (tensor<16xf32>, f32) -> tensor<16xf32>
  
  // Equivalent generic forms
  %2 = "func.constant"() { value = @myfn } : () -> ((tensor<16xf32>, f32) -> tensor<16xf32>)
  ```
  
  MLIR does not allow direct references to functions in SSA operands because
  the compiler is multithreaded, and disallowing SSA values to directly
  reference a function simplifies this
  ([rationale](../Rationale/Rationale.md#multithreading-the-compiler)).
  """

  OPERATION_NAME = "func.constant"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, value: _Union[str, _ods_ir.FlatSymbolRefAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["value"] = (value if (
    isinstance(value, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('FlatSymbolRefAttr')) else
      _ods_ir.AttrBuilder.get('FlatSymbolRefAttr')(value, context=_ods_context))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def value(self) -> _ods_ir.FlatSymbolRefAttr:
    return self.operation.attributes["value"]

  @value.setter
  def value(self, value: _ods_ir.FlatSymbolRefAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["value"] = value

@_ods_cext.register_op_adaptor(ConstantOp)
class ConstantOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "func.constant"

  @builtins.property
  def value(self) -> _ods_ir.FlatSymbolRefAttr:
    return self.attributes["value"]

def constant(result: _ods_ir.Type, value: _Union[str, _ods_ir.FlatSymbolRefAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return ConstantOp(result=result, value=value, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class FuncOp(_ods_ir.OpView):
  r"""
  Operations within the function cannot implicitly capture values defined
  outside of the function, i.e. Functions are `IsolatedFromAbove`. All
  external references must use function arguments or attributes that establish
  a symbolic connection (e.g. symbols referenced by name via a string
  attribute like SymbolRefAttr). An external function declaration (used when
  referring to a function declared in some other module) has no body. While
  the MLIR textual form provides a nice inline syntax for function arguments,
  they are internally represented as “block arguments” to the first block in
  the region.
  
  Only dialect attribute names may be specified in the attribute dictionaries
  for function arguments, results, or the function itself.
  
  Example:
  
  ```mlir
  // External function definitions.
  func.func private @abort()
  func.func private @scribble(i32, i64, memref<? x 128 x f32, #layout_map0>) -> f64
  
  // A function that returns its argument twice:
  func.func @count(%x: i64) -> (i64, i64)
    attributes {fruit = "banana"} {
    return %x, %x: i64, i64
  }
  
  // A function with an argument attribute
  func.func private @example_fn_arg(%x: i32 {swift.self = unit})
  
  // A function with a result attribute
  func.func private @example_fn_result() -> (f64 {dialectName.attrName = 0 : i64})
  
  // A function with an attribute
  func.func private @example_fn_attr() attributes {dialectName.attrName = false}
  ```
  """

  OPERATION_NAME = "func.func"

  _ODS_REGIONS = (1, True)

  def __init__(self, sym_name: _Union[str, _ods_ir.StringAttr], function_type: _Union[_Any, _ods_ir.TypeAttr], *, sym_visibility: _Optional[_Union[str, _ods_ir.StringAttr]] = None, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, no_inline: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["sym_name"] = (sym_name if (
    isinstance(sym_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolNameAttr')) else
      _ods_ir.AttrBuilder.get('SymbolNameAttr')(sym_name, context=_ods_context))
    attributes["function_type"] = (function_type if (
    isinstance(function_type, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('anonymous_452')) else
      _ods_ir.AttrBuilder.get('anonymous_452')(function_type, context=_ods_context))
    if sym_visibility is not None: attributes["sym_visibility"] = (sym_visibility if (
        isinstance(sym_visibility, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('StrAttr')) else
          _ods_ir.AttrBuilder.get('StrAttr')(sym_visibility, context=_ods_context))
    if arg_attrs is not None: attributes["arg_attrs"] = (arg_attrs if (
        isinstance(arg_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(arg_attrs, context=_ods_context))
    if res_attrs is not None: attributes["res_attrs"] = (res_attrs if (
        isinstance(res_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(res_attrs, context=_ods_context))
    if bool(no_inline): attributes["no_inline"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["sym_name"]

  @sym_name.setter
  def sym_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sym_name"] = value

  @builtins.property
  def function_type(self) -> _ods_ir.TypeAttr:
    return self.operation.attributes["function_type"]

  @function_type.setter
  def function_type(self, value: _ods_ir.TypeAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["function_type"] = value

  @builtins.property
  def sym_visibility(self) -> _Optional[_ods_ir.StringAttr]:
    if "sym_visibility" not in self.operation.attributes:
      return None
    return self.operation.attributes["sym_visibility"]

  @sym_visibility.setter
  def sym_visibility(self, value: _Optional[_ods_ir.StringAttr]):
    if value is not None:
      self.operation.attributes["sym_visibility"] = value
    elif "sym_visibility" in self.operation.attributes:
      del self.operation.attributes["sym_visibility"]

  @sym_visibility.deleter
  def sym_visibility(self):
    del self.operation.attributes["sym_visibility"]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["arg_attrs"]

  @arg_attrs.setter
  def arg_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["arg_attrs"] = value
    elif "arg_attrs" in self.operation.attributes:
      del self.operation.attributes["arg_attrs"]

  @arg_attrs.deleter
  def arg_attrs(self):
    del self.operation.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["res_attrs"]

  @res_attrs.setter
  def res_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["res_attrs"] = value
    elif "res_attrs" in self.operation.attributes:
      del self.operation.attributes["res_attrs"]

  @res_attrs.deleter
  def res_attrs(self):
    del self.operation.attributes["res_attrs"]

  @builtins.property
  def no_inline(self) -> bool:
    return "no_inline" in self.operation.attributes

  @no_inline.setter
  def no_inline(self, value):
    if bool(value):
      self.operation.attributes["no_inline"] = _ods_ir.UnitAttr.get()
    elif "no_inline" in self.operation.attributes:
      del self.operation.attributes["no_inline"]

  @no_inline.deleter
  def no_inline(self):
    del self.operation.attributes["no_inline"]

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(FuncOp)
class FuncOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "func.func"

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.attributes["sym_name"]

  @builtins.property
  def function_type(self) -> _ods_ir.TypeAttr:
    return self.attributes["function_type"]

  @builtins.property
  def sym_visibility(self) -> _Optional[_ods_ir.StringAttr]:
    if "sym_visibility" not in self.attributes:
      return None
    return self.attributes["sym_visibility"]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.attributes:
      return None
    return self.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.attributes:
      return None
    return self.attributes["res_attrs"]

  @builtins.property
  def no_inline(self) -> bool:
    return "no_inline" in self.attributes

def func(sym_name: _Union[str, _ods_ir.StringAttr], function_type: _Union[_Any, _ods_ir.TypeAttr], *, sym_visibility: _Optional[_Union[str, _ods_ir.StringAttr]] = None, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, no_inline: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> FuncOp:
  return FuncOp(sym_name=sym_name, function_type=function_type, sym_visibility=sym_visibility, arg_attrs=arg_attrs, res_attrs=res_attrs, no_inline=no_inline, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ReturnOp(_ods_ir.OpView):
  r"""
  The `func.return` operation represents a return operation within a function.
  The operation takes variable number of operands and produces no results.
  The operand number and types must match the signature of the function
  that contains the operation.
  
  Example:
  
  ```mlir
  func.func @foo() -> (i32, f8) {
    ...
    return %0, %1 : i32, f8
  }
  ```
  """

  OPERATION_NAME = "func.return"

  _ODS_REGIONS = (0, True)

  def __init__(self, operands_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(operands_))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(ReturnOp)
class ReturnOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "func.return"

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def return_(operands_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ReturnOp:
  return ReturnOp(operands_=operands_, loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import importlib.resources

__version__ = importlib.resources.files("numba_cuda_mlir").joinpath("VERSION").read_text().strip()




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class ApplyGPUPromoteShuffleToAMDGPUPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that are tryin to promote `gpu.shuffle`s to specialized
  AMDGPU intrinsics.
  """

  OPERATION_NAME = "transform.apply_patterns.gpu.gpu_shuffle_to_amdgpu"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, chipset: _Optional[_Union[str, _ods_ir.StringAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    if chipset is not None: attributes["chipset"] = (chipset if (
        isinstance(chipset, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('StrAttr')) else
          _ods_ir.AttrBuilder.get('StrAttr')(chipset, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def chipset(self) -> _Optional[_ods_ir.StringAttr]:
    if "chipset" not in self.operation.attributes:
      return None
    return self.operation.attributes["chipset"]

  @chipset.setter
  def chipset(self, value: _Optional[_ods_ir.StringAttr]):
    if value is not None:
      self.operation.attributes["chipset"] = value
    elif "chipset" in self.operation.attributes:
      del self.operation.attributes["chipset"]

  @chipset.deleter
  def chipset(self):
    del self.operation.attributes["chipset"]

@_ods_cext.register_op_adaptor(ApplyGPUPromoteShuffleToAMDGPUPatternsOp)
class ApplyGPUPromoteShuffleToAMDGPUPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_patterns.gpu.gpu_shuffle_to_amdgpu"

  @builtins.property
  def chipset(self) -> _Optional[_ods_ir.StringAttr]:
    if "chipset" not in self.attributes:
      return None
    return self.attributes["chipset"]

def apply_patterns_gpu_gpu_shuffle_to_amdgpu(*, chipset: _Optional[_Union[str, _ods_ir.StringAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyGPUPromoteShuffleToAMDGPUPatternsOp:
  return ApplyGPUPromoteShuffleToAMDGPUPatternsOp(chipset=chipset, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplyGPURewritePatternsOp(_ods_ir.OpView):
  r"""
  Collects GPU rewrite patterns comprising:
    1. GpuAllReduceRewrite patterns
    2. GpuGlobalIdRewriter patterns
    3. GpuShuffleRewriter patterns
  """

  OPERATION_NAME = "transform.apply_patterns.gpu.gpu_rewrite_patterns"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplyGPURewritePatternsOp)
class ApplyGPURewritePatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_patterns.gpu.gpu_rewrite_patterns"

def apply_patterns_gpu_gpu_rewrite_patterns(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyGPURewritePatternsOp:
  return ApplyGPURewritePatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplyGPUSubgroupReduceToNVVMConversionPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that convert GPU dialect ops related to wmma ops
  to NVVM dialect ops.
  These patterns require an "LLVMTypeConverter".
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_subgroup_reduce_to_nvvm"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplyGPUSubgroupReduceToNVVMConversionPatternsOp)
class ApplyGPUSubgroupReduceToNVVMConversionPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_subgroup_reduce_to_nvvm"

def apply_conversion_patterns_gpu_gpu_subgroup_reduce_to_nvvm(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyGPUSubgroupReduceToNVVMConversionPatternsOp:
  return ApplyGPUSubgroupReduceToNVVMConversionPatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplyGPUToNVVMConversionPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that convert GPU dialect ops to NVVM dialect ops. These
  patterns require an "LLVMTypeConverter".
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_to_nvvm"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, benefit: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    if benefit is not None: attributes["benefit"] = (benefit if (
        isinstance(benefit, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I16Attr')) else
          _ods_ir.AttrBuilder.get('I16Attr')(benefit, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def benefit(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["benefit"]

  @benefit.setter
  def benefit(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["benefit"] = value

@_ods_cext.register_op_adaptor(ApplyGPUToNVVMConversionPatternsOp)
class ApplyGPUToNVVMConversionPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_to_nvvm"

  @builtins.property
  def benefit(self) -> _ods_ir.IntegerAttr:
    return self.attributes["benefit"]

def apply_conversion_patterns_gpu_gpu_to_nvvm(*, benefit: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyGPUToNVVMConversionPatternsOp:
  return ApplyGPUToNVVMConversionPatternsOp(benefit=benefit, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplyGPUToROCDLConversionPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that convert GPU dialect ops to ROCDL dialect ops. These
  patterns require an "LLVMTypeConverter".
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_to_rocdl"

  _ODS_REGIONS = (0, True)

  def __init__(self, chipset: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["chipset"] = (chipset if (
    isinstance(chipset, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrAttr')) else
      _ods_ir.AttrBuilder.get('StrAttr')(chipset, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def chipset(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["chipset"]

  @chipset.setter
  def chipset(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["chipset"] = value

@_ods_cext.register_op_adaptor(ApplyGPUToROCDLConversionPatternsOp)
class ApplyGPUToROCDLConversionPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_to_rocdl"

  @builtins.property
  def chipset(self) -> _ods_ir.StringAttr:
    return self.attributes["chipset"]

def apply_conversion_patterns_gpu_gpu_to_rocdl(chipset: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyGPUToROCDLConversionPatternsOp:
  return ApplyGPUToROCDLConversionPatternsOp(chipset=chipset, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplyGPUWwmaToNVVMConversionPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that convert GPU dialect ops related to wmma ops
  to NVVM dialect ops.
  These patterns require an "LLVMTypeConverter".
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_wmma_to_nvvm"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplyGPUWwmaToNVVMConversionPatternsOp)
class ApplyGPUWwmaToNVVMConversionPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.gpu.gpu_wmma_to_nvvm"

def apply_conversion_patterns_gpu_gpu_wmma_to_nvvm(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyGPUWwmaToNVVMConversionPatternsOp:
  return ApplyGPUWwmaToNVVMConversionPatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplyUnrollVectorsSubgroupMmaOp(_ods_ir.OpView):
  r"""
  Unrolls contractions to the target `m`, `n`, and `k` native vector size,
  along with other vector operations based on expected usage. `transfer_read`
  ops unroll based on the extract slice shape introduced by unrolling the
  contractions, while elementwise and `transfer_write` ops unroll to the shape of
  the C matrix (`m x n`).
  
  This operation applies to pure vector operations and should be applied before
  lowering to subgroup_mma ops.
  """

  OPERATION_NAME = "transform.apply_patterns.gpu.unroll_vectors_subgroup_mma"

  _ODS_REGIONS = (0, True)

  def __init__(self, m: _Union[int, _ods_ir.IntegerAttr], n: _Union[int, _ods_ir.IntegerAttr], k: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["m"] = (m if (
    isinstance(m, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(m, context=_ods_context))
    attributes["n"] = (n if (
    isinstance(n, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(n, context=_ods_context))
    attributes["k"] = (k if (
    isinstance(k, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(k, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def m(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["m"]

  @m.setter
  def m(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["m"] = value

  @builtins.property
  def n(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["n"]

  @n.setter
  def n(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["n"] = value

  @builtins.property
  def k(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["k"]

  @k.setter
  def k(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["k"] = value

@_ods_cext.register_op_adaptor(ApplyUnrollVectorsSubgroupMmaOp)
class ApplyUnrollVectorsSubgroupMmaOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_patterns.gpu.unroll_vectors_subgroup_mma"

  @builtins.property
  def m(self) -> _ods_ir.IntegerAttr:
    return self.attributes["m"]

  @builtins.property
  def n(self) -> _ods_ir.IntegerAttr:
    return self.attributes["n"]

  @builtins.property
  def k(self) -> _ods_ir.IntegerAttr:
    return self.attributes["k"]

def apply_patterns_gpu_unroll_vectors_subgroup_mma(m: _Union[int, _ods_ir.IntegerAttr], n: _Union[int, _ods_ir.IntegerAttr], k: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyUnrollVectorsSubgroupMmaOp:
  return ApplyUnrollVectorsSubgroupMmaOp(m=m, n=n, k=k, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class EliminateBarriersOp(_ods_ir.OpView):
  r"""
  Removes unnecessary GPU barriers from the function. If a barrier does not
  enforce any conflicting pair of memory effects, including a pair that is
  enforced by another barrier, it is unnecessary and can be removed.
  
  The approach is based on "High-Performance GPU-to-CPU Transpilation and
  Optimization via High-Level Parallel Constructs" by  Moses, Ivanov,
  Domke, Endo, Doerfert, and Zinenko in PPoPP 2023. Specifically, it
  analyzes the memory effects of the operations before and after the given
  barrier and checks if the barrier enforces any of the memory
  effect-induced dependencies that aren't already enforced by another
  barrier.
  
  For example, in the following code
  
  ```mlir
    store %A
    barrier  // enforces load-after-store
    load %A
    barrier  // load-after-store already enforced by the previous barrier
    load %A
  ```
  
  the second barrier can be removed.
  """

  OPERATION_NAME = "transform.apply_patterns.gpu.eliminate_barriers"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(EliminateBarriersOp)
class EliminateBarriersOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_patterns.gpu.eliminate_barriers"

def apply_patterns_gpu_eliminate_barriers(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> EliminateBarriersOp:
  return EliminateBarriersOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class MapForallToBlocks(_ods_ir.OpView):
  r"""
  Target the gpu_launch op and rewrite the top level `scf.forall`
  to distributed gpu.block_id attribute. If `generate_gpu_launch` attribute
  is set, then first generates `gpu_launch` and moves the top level
  `scf.forall` inside.
  
  The operation searches top level `scf.forall` ops under
  `gpu_launch` and maps each such op to GPU blocks. Mapping is
  one-to-one and the induction variables of `scf.forall` are
  rewritten to gpu.block_id according to the `thread_dim_mapping` attribute.
  
  Dynamic, `scf.forall` trip counts are currently not supported.
  Dynamic block dim sizes are currently not supported.
  
  Only **bufferized** scf.forall are currently supported.
  Only scf.forall distributed to **at most 3 dimensions** are
  currently supported.
  
  The operation alters the block size of the given gpu_launch using the
  grid_dims argument.
  
  #### Return modes:
  
  This operation ignores non-gpu_launch ops and drops them in the return.
  
  If any scf.forall with tensors is found, the transform definitely
  fails.
  
  If all the `scf.forall` operations contained within the LaunchOp
  referred to by the `target` handle lower to GPU properly, the
  transform succeeds. Otherwise the transform definitely fails.
  
  The returned handle points to the same LaunchOp operand, consuming it and
  producing a new SSA value to satisfy chaining and linearity of the IR
  properties.
  """

  OPERATION_NAME = "transform.gpu.map_forall_to_blocks"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, target: _ods_ir.Value, *, grid_dims: _Optional[_Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr]] = None, generate_gpu_launch: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if grid_dims is not None: attributes["grid_dims"] = (grid_dims if (
        isinstance(grid_dims, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DenseI64ArrayAttr')) else
          _ods_ir.AttrBuilder.get('DenseI64ArrayAttr')(grid_dims, context=_ods_context))
    if bool(generate_gpu_launch): attributes["generate_gpu_launch"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def grid_dims(self) -> _Optional[_ods_ir.DenseI64ArrayAttr]:
    if "grid_dims" not in self.operation.attributes:
      return None
    return self.operation.attributes["grid_dims"]

  @grid_dims.setter
  def grid_dims(self, value: _Optional[_ods_ir.DenseI64ArrayAttr]):
    if value is not None:
      self.operation.attributes["grid_dims"] = value
    elif "grid_dims" in self.operation.attributes:
      del self.operation.attributes["grid_dims"]

  @grid_dims.deleter
  def grid_dims(self):
    del self.operation.attributes["grid_dims"]

  @builtins.property
  def generate_gpu_launch(self) -> bool:
    return "generate_gpu_launch" in self.operation.attributes

  @generate_gpu_launch.setter
  def generate_gpu_launch(self, value):
    if bool(value):
      self.operation.attributes["generate_gpu_launch"] = _ods_ir.UnitAttr.get()
    elif "generate_gpu_launch" in self.operation.attributes:
      del self.operation.attributes["generate_gpu_launch"]

  @generate_gpu_launch.deleter
  def generate_gpu_launch(self):
    del self.operation.attributes["generate_gpu_launch"]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MapForallToBlocks)
class MapForallToBlocksAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.gpu.map_forall_to_blocks"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def grid_dims(self) -> _Optional[_ods_ir.DenseI64ArrayAttr]:
    if "grid_dims" not in self.attributes:
      return None
    return self.attributes["grid_dims"]

  @builtins.property
  def generate_gpu_launch(self) -> bool:
    return "generate_gpu_launch" in self.attributes

def gpu_map_forall_to_blocks(result: _ods_ir.Type, target: _ods_ir.Value, *, grid_dims: _Optional[_Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr]] = None, generate_gpu_launch: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MapForallToBlocks(result=result, target=target, grid_dims=grid_dims, generate_gpu_launch=generate_gpu_launch, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MapNestedForallToThreads(_ods_ir.OpView):
  r"""
  Target the `gpu.launch op` and rewrite all `scf.forall` nested in it to
  distributed `gpu.thread_id` attribute.
  
  The operation searches for `scf.forall` ops nested under `target` and maps
  each such op to GPU threads.
  
  `scf.forall` induction variables are rewritten to `gpu.thread_id` according
  to the `mapping` attribute.
  
  Different types of mappings attributes are supported:
    - the block_dims is a list of integers that specifies the number of
      threads in each dimension. This is a mandatory attribute that is used
      to constrain the number of threads in each dimension. If an
      `scf.forall` op is mapped to fewer threads, predication occurs.
    - the warp_dims is a list of integers that specifies the number of
      warps in each dimension. This is an optional attribute that is used
      to constrain the number of warps in each dimension. When present, this
      attribute must be specified in a way that is compatible with the
      block_dims attribute. If an `scf.forall` op is mapped to fewer warps,
      predication occurs.
  
  Dynamic `scf.forall` trip counts are currently not supported.
  Dynamic block dim sizes are currently not supported.
  
  Only **bufferized** `scf.forall` are currently supported.
  Only `scf.forall` distributed to **at most 3 dimensions** are
  currently supported.
  
  The `sync_after_distribute`attribute controls whether a `gpu.barrier` is
  inserted after each scf.forall op. At this time, this is an all or nothing
  choice. This will need to be tightened in the future.
  
  The operation alters the block size of the given gpu_launch using the
  mandatory block_dims argument.
  
  #### Return modes:
  
  This operation ignores non-`gpu_launch` ops and drops them in the return.
  
  If any scf.forall with tensors is found, the transform definitely
  fails.
  
  If all the `scf.forall` operations with gpu.thread mapping contained
  within the `LaunchOp` referred to by the `target` handle lower to GPU
  properly, the transform succeeds. Otherwise the transform definitely
  fails.
  
  scf.forall operations with mappings other than gpu.thread are
  ignored.
  
  The returned handle points to the same LaunchOp operand, consuming it and
  producing a new SSA value to satisfy chaining and linearity of the IR
  properties.
  
  #### Example:
  
  ```
  gpu.launch blocks(%bx, %by, %bz) in (%x = %0, %y = %1, %z = %2)
             threads(%tx, %ty, %tz) in (%tx = %3, %ty = %4, %tz = %5) {
    scf.forall (%i, %j) in (7, 9) {
      ... // body 1
    } {mapping = [#gpu.thread<x>, #gpu.thread<y>, #gpu.thread<z>]}
    scf.forall (%i) in (12) {
      ... // body 2
    } {mapping = [#gpu.thread<x>]}
    gpu.terminator
  }
  ```
  
  is translated to:
  
  ```
  %bdimX = arith.constant 12 : index
  %bdimY = arith.constant 9 : index
  gpu.launch blocks(%bx, %by, %bz) in (%x = %0, %y = %1, %z = %2)
         threads(%tx, %ty, %tz) in (%tx = %bdimX, %ty = %bdimY, %tz = %5) {
    if (threadIdx.x < 9 && threadIdx.y < 7) {
      ... // body 1
    }
    gpu.barrier
    if (threadIdx.y < 1) {
      ... // body 2
    }
    gpu.barrier
    gpu.terminator
  }
  ```
  """

  OPERATION_NAME = "transform.gpu.map_nested_forall_to_threads"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, target: _ods_ir.Value, *, block_dims: _Optional[_Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr]] = None, sync_after_distribute: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, warp_size: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if block_dims is not None: attributes["block_dims"] = (block_dims if (
        isinstance(block_dims, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DenseI64ArrayAttr')) else
          _ods_ir.AttrBuilder.get('DenseI64ArrayAttr')(block_dims, context=_ods_context))
    if sync_after_distribute is not None: attributes["sync_after_distribute"] = (sync_after_distribute if (
        isinstance(sync_after_distribute, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(sync_after_distribute, context=_ods_context))
    if warp_size is not None: attributes["warp_size"] = (warp_size if (
        isinstance(warp_size, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I64Attr')) else
          _ods_ir.AttrBuilder.get('I64Attr')(warp_size, context=_ods_context))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def block_dims(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.operation.attributes["block_dims"]

  @block_dims.setter
  def block_dims(self, value: _ods_ir.DenseI64ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["block_dims"] = value

  @builtins.property
  def sync_after_distribute(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["sync_after_distribute"]

  @sync_after_distribute.setter
  def sync_after_distribute(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sync_after_distribute"] = value

  @builtins.property
  def warp_size(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["warp_size"]

  @warp_size.setter
  def warp_size(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["warp_size"] = value

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MapNestedForallToThreads)
class MapNestedForallToThreadsAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.gpu.map_nested_forall_to_threads"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def block_dims(self) -> _ods_ir.DenseI64ArrayAttr:
    return self.attributes["block_dims"]

  @builtins.property
  def sync_after_distribute(self) -> _ods_ir.BoolAttr:
    return self.attributes["sync_after_distribute"]

  @builtins.property
  def warp_size(self) -> _ods_ir.IntegerAttr:
    return self.attributes["warp_size"]

def gpu_map_nested_forall_to_threads(result: _ods_ir.Type, target: _ods_ir.Value, *, block_dims: _Optional[_Union[_Sequence[int], _ods_ir.DenseI64ArrayAttr]] = None, sync_after_distribute: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, warp_size: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MapNestedForallToThreads(result=result, target=target, block_dims=block_dims, sync_after_distribute=sync_after_distribute, warp_size=warp_size, loc=loc, ip=ip).result



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.types import int64, Type
from .aggregate_types import AggregateType


class CUTensorMapStorageType(AggregateType):
    """
    Tensor map descriptor. See the cuda declaration:

    /**
    * Tensor map descriptor. Requires compiler support for aligning to 64 bytes.
    */
    typedef struct CUtensorMap_st {
        alignas(64)
        cuuint64_t opaque[CU_TENSOR_MAP_NUM_QWORDS];
    } CUtensorMap;
    """

    CU_TENSOR_MAP_NUM_QWORDS = 16

    def __init__(self):
        super().__init__(
            "CUTensorMapStorage",
            [(f"f{i}", int64) for i in range(self.CU_TENSOR_MAP_NUM_QWORDS)],
        )


class CUTensorMapType(Type):
    """
    A byval(CUTensorMapStorage) type.
    """

    def __init__(self):
        super().__init__("CUTensorMap")
        self.__numba_cuda_mlir_attributes__ = {"llvm.byval": CUTensorMapStorageType()}



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import lru_cache
import inspect
from numba_cuda_mlir.errors import ForceLiteralArg
from numba_cuda_mlir.numba_cuda.typing import typeof
from numba_cuda_mlir.numba_cuda.misc.special import literal_unroll
from numba_cuda_mlir.numba_cuda.misc.special import literally
from numba_cuda_mlir.numba_cuda.misc.special import (
    literal_unroll as cuda_literal_unroll,
)
from numba_cuda_mlir.numba_cuda.typing.templates import (
    CallableTemplate,
    AttributeTemplate,
    ConcreteTemplate,
    AbstractTemplate,
    Registry,
    signature,
)
from numba_cuda_mlir.numba_cuda.typing.npydecl import parse_dtype, parse_shape
from numba_cuda_mlir import types
from numba_cuda_mlir.cuda.vector_types import (
    vector_types_by_name,
    vector_types_by_alias,
)

from numba_cuda_mlir.numba_cuda import cudadecl

registry = Registry()


class CArrayTyping(AbstractTemplate):
    from numba_cuda_mlir.cuda import carray

    key = carray
    layout = "C"

    def generic(self, args, kws):
        if kws:
            if set(kws) != {"dtype"}:
                return None
            args = (*args, kws["dtype"])

        if len(args) not in (2, 3):
            return None

        ptr, shape = args[:2]
        dtype_arg = args[2] if len(args) == 3 else None

        if ptr is types.voidptr:
            ptr_dtype = None
        elif isinstance(ptr, types.CPointer):
            ptr_dtype = ptr.dtype
        else:
            return None

        if dtype_arg is None or isinstance(dtype_arg, types.NoneType):
            if ptr_dtype is None:
                return None
            dtype = ptr_dtype
        else:
            dtype = parse_dtype(dtype_arg)
            if dtype is None or (ptr_dtype is not None and dtype != ptr_dtype):
                return None

        ndim = parse_shape(shape)
        if ndim is None:
            return None

        return signature(types.Array(dtype, ndim, self.layout), *args)


registry.register_global(CArrayTyping.key, types.Function(CArrayTyping))


class FArrayTyping(CArrayTyping):
    from numba_cuda_mlir.cuda import farray

    key = farray
    layout = "F"


registry.register_global(FArrayTyping.key, types.Function(FArrayTyping))


@registry.register
class PrintFunctionTemplate(AbstractTemplate):
    from numba_cuda_mlir.cuda.print import print as print_intrinsic

    key = print_intrinsic

    def is_valid_print_argument_type(self, arg_type):
        from numba_cuda_mlir.numba_cuda.types.ext_types import Dim3

        if isinstance(arg_type, types.BaseTuple):
            return all(self.is_valid_print_argument_type(t) for t in arg_type.types)
        return isinstance(
            arg_type,
            (
                types.Number,
                types.IntegerLiteral,
                types.Boolean,
                types.Array,
                types.UnicodeType,
                types.StringLiteral,
                Dim3,
            ),
        )

    def generic(self, args, kws):
        for arg in args:
            if not self.is_valid_print_argument_type(arg):
                raise TypeError(f"Invalid argument type for print: {arg=} ({type(arg)=})")

        if "end" in kws:
            end = kws.pop("end")
            if not isinstance(end, (types.StringLiteral, types.UnicodeType)):
                raise TypeError(f"Invalid end type for print: {end=} ({type(end)=})")

        if "sep" in kws:
            sep = kws.pop("sep")
            if not isinstance(sep, (types.StringLiteral, types.UnicodeType)):
                raise TypeError(f"Invalid sep type for print: {sep=} ({type(sep)=})")

        return signature(types.none, *args)


@registry.register
class BuiltinPrintFunctionTemplate(PrintFunctionTemplate):
    key = print


class InlinePTXFunctionTemplate(AbstractTemplate):
    from numba_cuda_mlir.cuda import inline_ptx

    key = inline_ptx

    def generic(self, args, kws):
        from numba_cuda_mlir.lowering_utilities.type_conversions import (
            inline_ptx_type_constraint_to_numba_type,
        )

        if len(args) == 0:
            raise TypeError(f"Invalid arguments to inline_ptx: {args=}")
        force_args_to_be_literal = set()

        format, *rest = args
        if not isinstance(format, types.Literal):
            force_args_to_be_literal.add(0)

        if len(rest) % 2 != 0:
            raise TypeError(f"Invalid arguments to inline_ptx: {args=}")

        result_types = []
        arg_types = [types.string]

        for i, (fmt, _arg) in enumerate(zip(rest[::2], rest[1::2])):
            if not isinstance(fmt, types.Literal):
                force_args_to_be_literal.add(1 + i * 2)
            else:
                fmt_string = fmt.literal_value
                arg_type = inline_ptx_type_constraint_to_numba_type(fmt_string[-1])
                arg_types.extend([fmt, arg_type])
                if "=" in fmt_string:
                    # write-only arguments are returned
                    result_types.append(arg_type)

        if len(force_args_to_be_literal) > 0:
            from numba_cuda_mlir.numba_cuda.core.errors import ForceLiteralArg

            raise ForceLiteralArg(force_args_to_be_literal)

        if len(result_types) == 0:
            result_type = types.none
        elif len(result_types) == 1:
            # Single output: return scalar type
            result_type = result_types[0]
        else:
            # Multiple outputs: return tuple
            result_type = types.Tuple(result_types)
        return signature(result_type, *arg_types)


@registry.register
class PointerCastTemplate(AbstractTemplate):
    from numba_cuda_mlir.types import ptr

    key = ptr

    def generic(self, args, kws):
        if len(args) != 1:
            raise TypeError(f"ptr() takes exactly 1 argument ({len(args)} given)")
        if isinstance(args[0], (types.Array, types.CPointer, types.Integer, types.AggregateType)):
            return signature(types.CPointer(types.none), args[0])
        else:
            raise TypeError(f"Invalid argument type for pointer cast: {args[0]=}")


class SyncthreadsTemplate(ConcreteTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.syncthreads
    cases = [types.none()]


class NumbaSyncthreadsTemplate(SyncthreadsTemplate):
    from numba_cuda_mlir.numba_cuda import syncthreads

    key = syncthreads


class ThisGridTemplate(ConcreteTemplate):
    from numba_cuda_mlir.numba_cuda.cg import this_grid
    from numba_cuda_mlir.numba_cuda.types.ext_types import grid_group

    key = this_grid
    cases = [signature(grid_group)]


class GridGroupSyncTemplate(AbstractTemplate):
    key = "GridGroup.sync"

    def generic(self, args, kws):
        if len(args) == 0:
            return signature(types.int32, recvr=self.this)
        return None


class ShflSyncTemplateBase(AbstractTemplate):
    def generic(self, args, kws):
        if len(args) != 3:
            return None
        mask, value, src_lane = args
        if not isinstance(mask, types.Integer):
            return None
        if not isinstance(value, (types.Integer, types.Float)):
            return None
        if not isinstance(src_lane, types.Integer):
            return None
        return signature(value, mask, value, src_lane)


class ShflSyncTemplate(ShflSyncTemplateBase):
    from numba_cuda_mlir import cuda

    key = cuda.shfl_sync


class ShflUpSyncTemplate(ShflSyncTemplateBase):
    from numba_cuda_mlir import cuda

    key = cuda.shfl_up_sync


class ShflDownSyncTemplate(ShflSyncTemplateBase):
    from numba_cuda_mlir import cuda

    key = cuda.shfl_down_sync


class ShflXorSyncTemplate(ShflSyncTemplateBase):
    from numba_cuda_mlir import cuda

    key = cuda.shfl_xor_sync


class GenericArrayTemplate(CallableTemplate):
    # Subclasses can set this to True to allow non-literal shapes (e.g., for dynamic shared memory)
    allow_dynamic_shape = False

    def generic(self):
        allow_dynamic = self.allow_dynamic_shape

        def typer(shape, dtype, alignment=None, alignas=None):
            # By default, only integer literals and tuples of integer literals are valid shapes
            # This matches numba-cuda's Cuda_array_decl behavior for static allocations.
            # Shared arrays can use dynamic shapes when allow_dynamic_shape=True.
            if isinstance(shape, types.Integer):
                if not allow_dynamic and not isinstance(shape, types.IntegerLiteral):
                    return None
                ndim = 1
            elif isinstance(shape, (types.Tuple, types.UniTuple)):
                if not allow_dynamic and any(
                    not isinstance(s, types.IntegerLiteral) for s in shape
                ):
                    return None
                ndim = len(shape)
            else:
                return None

            # Support both 'alignment' (numba-cuda) and 'alignas' (numba_cuda_mlir) keywords
            align_val = alignment if alignment is not None else alignas

            if align_val is not None:
                permitted = (types.IntegerLiteral, types.NoneType)
                if not isinstance(align_val, permitted):
                    from numba_cuda_mlir.numba_cuda.core.errors import (
                        RequireLiteralValue,
                    )

                    raise RequireLiteralValue("alignment must be a constant integer")

            # Parse dtype - handle DTypeSpec, TypeRef, and StringLiteral
            if isinstance(dtype, types.DTypeSpec):
                nb_dtype = dtype.dtype
            elif isinstance(dtype, types.TypeRef):
                nb_dtype = dtype.instance_type
            elif isinstance(dtype, types.StringLiteral):
                import numpy as np
                from numba_cuda_mlir.numba_cuda.core.errors import TypingError
                from numba_cuda_mlir.numba_cuda.np.numpy_support import from_dtype

                try:
                    dt = np.dtype(dtype.literal_value)
                except TypeError:
                    raise TypingError(f"Invalid NumPy dtype specified: '{dtype.literal_value}'")
                nb_dtype = from_dtype(dt)
            else:
                return None

            if nb_dtype is not None and ndim is not None:
                return types.Array(dtype=nb_dtype, ndim=ndim, layout="C")

            return None

        return typer


class ConstArrayLikeTemplate(AbstractTemplate):
    import numba_cuda_mlir.cuda

    key = numba_cuda_mlir.cuda.const.array_like

    def generic(self, args, kws):
        if len(args) == 1 and isinstance(args[0], types.Array):
            arr = args[0]
            # const.array_like returns an array with same shape and dtype
            return signature(types.Array(dtype=arr.dtype, ndim=arr.ndim, layout=arr.layout), arr)
        return None


@registry.register
class ConstArrayTemplate(ConstArrayLikeTemplate):
    pass


@registry.register
class LocalArrayTemplate(GenericArrayTemplate):
    import numba_cuda_mlir.cuda

    key = numba_cuda_mlir.cuda.local.array


@registry.register_attr
class CUDALocalModule(AttributeTemplate):
    import numba_cuda_mlir.cuda.local as local

    key = local

    def resolve_array(self, mod):
        return types.Function(LocalArrayTemplate)


@registry.register
class SharedArrayTemplate(GenericArrayTemplate):
    import numba_cuda_mlir.cuda

    key = numba_cuda_mlir.cuda.shared.array
    # Enable dynamic shapes for shared arrays (uses extern/dynamic shared memory at runtime)
    allow_dynamic_shape = True


@registry.register_attr
class CUDASharedModule(AttributeTemplate):
    import numba_cuda_mlir.cuda

    key = numba_cuda_mlir.cuda.shared

    def resolve_array(self, mod):
        return types.Function(SharedArrayTemplate)


@registry.register_attr
class CudaSharedModuleTemplate(AttributeTemplate):
    import numba_cuda_mlir.cuda

    key = types.Module(numba_cuda_mlir.cuda.shared)

    def resolve_array(self, mod):
        return types.Function(SharedArrayTemplate)


@registry.register_attr
class CudaConstModuleTemplate(AttributeTemplate):
    import numba_cuda_mlir.cuda

    key = types.Module(numba_cuda_mlir.cuda.const)

    def resolve_array_like(self, mod):
        return types.Function(ConstArrayTemplate)


@registry.register_attr
class CudaLocalModuleTemplate(AttributeTemplate):
    import numba_cuda_mlir.cuda

    key = types.Module(numba_cuda_mlir.cuda.local)

    def resolve_array(self, mod):
        return types.Function(LocalArrayTemplate)


@registry.register_attr
class CudaFP16ModuleTemplate(AttributeTemplate):
    from numba_cuda_mlir.cuda import fp16

    key = types.Module(fp16)

    @lru_cache
    def resolve(self, typ: types.Module, attr: str):
        func = getattr(typ.pymod, attr)
        sig = inspect.signature(func)
        params = sig.parameters.values()
        param_types = [param.annotation for param in params]
        return_type = sig.return_annotation

        class FP16FunctionTemplate(ConcreteTemplate):
            key = func
            cases = [signature(return_type, *param_types)]

        return types.Function(FP16FunctionTemplate)


@registry.register
class CudaGridTemplate(AbstractTemplate):
    import numba_cuda_mlir.cuda

    key = numba_cuda_mlir.cuda.grid

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 1
        if not isinstance(args[0], types.Literal):
            raise ForceLiteralArg(arg_indices={0})
        if not isinstance(args[0], types.Integer):
            return None
        value = args[0].literal_value
        if value == 1:
            return signature(types.int32, types.int32)
        ret_types = [types.int32 for i in range(value)]
        ret_ty = types.Tuple(ret_types)
        return signature(ret_ty, types.int32)


@registry.register
class CudaGridsizeTemplate(CudaGridTemplate):
    import numba_cuda_mlir.cuda

    key = numba_cuda_mlir.cuda.gridsize


class BitIntrinsicTemplate(AbstractTemplate):
    """Base template for bit manipulation intrinsics (clz, ffs, popc).

    These intrinsics take an integer and return int32.
    The input is NOT widened - operation happens at the input's bitwidth.
    """

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arg = args[0]
        if not isinstance(arg, types.Integer):
            return None
        return signature(types.int32, arg)


class ClzTemplate(BitIntrinsicTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.clz


class FfsTemplate(BitIntrinsicTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.ffs


class BrevTemplate(AbstractTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.brev

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arg = args[0]
        if not isinstance(arg, types.Integer):
            return None
        # brev returns the same type as input (reversed bits)
        return signature(arg, arg)


class PopcTemplate(BitIntrinsicTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.popc


class SelpTemplate(AbstractTemplate):
    """Select based on predicate: returns a if cond is true, else b."""

    from numba_cuda_mlir import cuda

    key = cuda.selp

    def generic(self, args, kws):
        if len(args) != 3:
            return None
        cond, a, b = args
        # cond must be integer or boolean
        if not isinstance(cond, (types.Integer, types.Boolean)):
            return None
        # a and b must be same type (or compatible)
        if isinstance(a, types.Integer) and isinstance(b, types.Integer):
            # Return the wider type
            if a.bitwidth >= b.bitwidth:
                return signature(a, cond, a, b)
            else:
                return signature(b, cond, a, b)
        elif isinstance(a, types.Float) and isinstance(b, types.Float):
            if a.bitwidth >= b.bitwidth:
                return signature(a, cond, a, b)
            else:
                return signature(b, cond, a, b)
        elif type(a) == type(b):
            return signature(a, cond, a, b)
        return None


def register_bit_intrinsics():
    from numba_cuda_mlir import cuda

    registry.register_global(cuda.clz)(ClzTemplate)
    registry.register_global(cuda.ffs)(FfsTemplate)
    registry.register_global(cuda.brev)(BrevTemplate)
    registry.register_global(cuda.popc)(PopcTemplate)
    registry.register_global(cuda.selp)(SelpTemplate)


register_bit_intrinsics()


@registry.register_attr
class Cuda_stub_resolver(cudadecl.CudaModuleTemplate, AttributeTemplate):
    import numba_cuda_mlir.cuda

    key = types.Module(numba_cuda_mlir.cuda)

    def resolve_grid(self, mod):
        return types.Function(CudaGridTemplate)

    def resolve_gridsize(self, mod):
        return types.Function(CudaGridsizeTemplate)

    def resolve_fp16(self, mod):
        from numba_cuda_mlir.cuda import fp16

        return types.Module(fp16)

    def resolve_local(self, mod):
        import numba_cuda_mlir.cuda

        return types.Module(numba_cuda_mlir.cuda.local)

    def resolve_syncthreads(self, mod):
        return types.Function(SyncthreadsTemplate)

    def resolve_shared(self, mod):
        import numba_cuda_mlir.cuda

        return types.Module(numba_cuda_mlir.cuda.shared)

    def resolve_const(self, mod):
        import numba_cuda_mlir.cuda

        return types.Module(numba_cuda_mlir.cuda.const)

    def resolve_shared_array(self, mod):
        return types.Function(SharedArrayTemplate)

    def resolve_local_array(self, mod):
        return types.Function(LocalArrayTemplate)

    def resolve_experimental(self, mod):
        import numba_cuda_mlir.cuda.experimental as experimental

        return types.Module(experimental)

    def resolve_intrin(self, mod):
        import numba_cuda_mlir.cuda.intrin as intrin

        return types.Module(intrin)

    def resolve_print(self, mod):
        return types.Function(PrintFunctionTemplate)

    def resolve_libdevice(self, mod):
        import numba_cuda_mlir.cuda.libdevice as libdevice

        return types.Module(libdevice)

    def resolve_shfl_sync(self, mod):
        return types.Function(ShflSyncTemplate)

    def resolve_shfl_up_sync(self, mod):
        return types.Function(ShflUpSyncTemplate)

    def resolve_shfl_down_sync(self, mod):
        return types.Function(ShflDownSyncTemplate)

    def resolve_shfl_xor_sync(self, mod):
        return types.Function(ShflXorSyncTemplate)

    def resolve_clz(self, mod):
        return types.Function(ClzTemplate)

    def resolve_ffs(self, mod):
        return types.Function(FfsTemplate)

    def resolve_brev(self, mod):
        return types.Function(BrevTemplate)

    def resolve_popc(self, mod):
        return types.Function(PopcTemplate)

    def resolve_selp(self, mod):
        return types.Function(SelpTemplate)

    def resolve_warpsize(self, mod):
        return types.int32

    def resolve_laneid(self, mod):
        return types.int32

    def resolve_all_sync(self, mod):
        return types.Function(AllSyncTemplate)

    def resolve_any_sync(self, mod):
        return types.Function(AnySyncTemplate)

    def resolve_eq_sync(self, mod):
        return types.Function(EqSyncTemplate)

    def resolve_ballot_sync(self, mod):
        return types.Function(BallotSyncTemplate)

    def resolve_match_any_sync(self, mod):
        return types.Function(MatchAnySyncTemplate)

    def resolve_match_all_sync(self, mod):
        return types.Function(MatchAllSyncTemplate)

    def resolve_activemask(self, mod):
        return types.Function(ActivemaskTemplate)

    def resolve_ldca(self, mod):
        return types.Function(LdcaTemplate)

    def resolve_ldcg(self, mod):
        return types.Function(LdcgTemplate)

    def resolve_ldcs(self, mod):
        return types.Function(LdcsTemplate)

    def resolve_ldlu(self, mod):
        return types.Function(LdluTemplate)

    def resolve_ldcv(self, mod):
        return types.Function(LdcvTemplate)

    def resolve_stcg(self, mod):
        return types.Function(StcgTemplate)

    def resolve_stcs(self, mod):
        return types.Function(StcsTemplate)

    def resolve_stwb(self, mod):
        return types.Function(StwbTemplate)

    def resolve_stwt(self, mod):
        return types.Function(StwtTemplate)

    def resolve_vector(self, mod):
        from numba_cuda_mlir.cuda import vector

        return types.Module(vector)

    def resolve_cg(self, mod):
        import numba_cuda_mlir.numba_cuda.cg as cg

        return types.Module(cg)

    def generic_resolve(self, mod, attr):
        from numba_cuda_mlir.numba_cuda.typing.typeof import typeof

        # Handle vector type constructors (float64x4, float32x2, int32x4, etc.)
        stub = vector_types_by_name.get(attr) or vector_types_by_alias.get(attr)
        if stub is not None:
            return typeof(stub)

        return None


@registry.register_attr
class CudaModuleTemplate(Cuda_stub_resolver):
    from numba_cuda_mlir import cuda

    key = types.Module(cuda)


@registry.register_attr
class RealNumbaCudaModuleTemplate(Cuda_stub_resolver):
    import numba_cuda_mlir.numba_cuda as numba_cuda_module

    key = types.Module(numba_cuda_module)


@registry.register_attr
class CudaExperimentalModuleTemplate(AttributeTemplate):
    import numba_cuda_mlir.cuda.experimental as experimental

    key = types.Module(experimental)

    def resolve_inline_ptx(self, mod):
        return types.Function(InlinePTXFunctionTemplate)

    def resolve_intrin(self, mod):
        import numba_cuda_mlir.cuda.intrin as intrin

        return types.Module(intrin)


@registry.register_attr
class CgModuleTemplate(AttributeTemplate):
    import numba_cuda_mlir.numba_cuda.cg as cg

    key = types.Module(cg)

    def resolve_this_grid(self, mod):
        return types.Function(ThisGridTemplate)


@registry.register_attr
class GridGroupAttributeTemplate(AttributeTemplate):
    from numba_cuda_mlir.numba_cuda.types.ext_types import GridGroup as GridGroupClass

    key = GridGroupClass

    def resolve_sync(self, grid_group_type):
        return types.BoundFunction(GridGroupSyncTemplate, grid_group_type)


def register_syncthreads_variants():
    from numba_cuda_mlir import cuda
    from numba_cuda_mlir._mlir.dialects import nvvm

    for intrin, _reduction_op in [
        (cuda.syncthreads_and, nvvm.BarrierReduction.AND),
        (cuda.syncthreads_or, nvvm.BarrierReduction.OR),
        (cuda.syncthreads_count, nvvm.BarrierReduction.POPC),
    ]:

        class STTy(ConcreteTemplate):
            cases = [types.i4(types.i4)]

        registry.register_global(intrin)(STTy)

    registry.register_global(cuda.syncthreads)(SyncthreadsTemplate)


register_syncthreads_variants()


def register_shfl_sync_intrinsics():
    from numba_cuda_mlir import cuda

    registry.register_global(cuda.shfl_sync)(ShflSyncTemplate)
    registry.register_global(cuda.shfl_up_sync)(ShflUpSyncTemplate)
    registry.register_global(cuda.shfl_down_sync)(ShflDownSyncTemplate)
    registry.register_global(cuda.shfl_xor_sync)(ShflXorSyncTemplate)


register_shfl_sync_intrinsics()


class VoteSyncPredicateTemplate(AbstractTemplate):
    """Template for vote_sync operations that return a boolean predicate (all_sync, any_sync, eq_sync)."""

    def generic(self, args, kws):
        if len(args) != 2:
            return None
        mask, predicate = args
        if not isinstance(mask, types.Integer):
            return None
        return signature(types.boolean, mask, predicate)


class AllSyncTemplate(VoteSyncPredicateTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.all_sync


class AnySyncTemplate(VoteSyncPredicateTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.any_sync


class EqSyncTemplate(VoteSyncPredicateTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.eq_sync


class BallotSyncTemplate(AbstractTemplate):
    """Template for ballot_sync which returns a mask of threads."""

    from numba_cuda_mlir import cuda

    key = cuda.ballot_sync

    def generic(self, args, kws):
        if len(args) != 2:
            return None
        mask, predicate = args
        if not isinstance(mask, types.Integer):
            return None
        return signature(types.uint32, mask, predicate)


class MatchAnySyncTemplate(AbstractTemplate):
    """Template for match_any_sync which returns a mask of threads with same value."""

    from numba_cuda_mlir import cuda

    key = cuda.match_any_sync

    def generic(self, args, kws):
        if len(args) != 2:
            return None
        mask, value = args
        if not isinstance(mask, types.Integer):
            return None
        if not isinstance(value, (types.Integer, types.Float)):
            return None
        return signature(types.uint32, mask, value)


class MatchAllSyncTemplate(AbstractTemplate):
    """Template for match_all_sync which returns (mask, predicate) tuple."""

    from numba_cuda_mlir import cuda

    key = cuda.match_all_sync

    def generic(self, args, kws):
        if len(args) != 2:
            return None
        mask, value = args
        if not isinstance(mask, types.Integer):
            return None
        if not isinstance(value, (types.Integer, types.Float)):
            return None
        return signature(types.Tuple([types.uint32, types.boolean]), mask, value)


class ActivemaskTemplate(AbstractTemplate):
    """Template for activemask which returns the mask of active threads."""

    from numba_cuda_mlir import cuda

    key = cuda.activemask

    def generic(self, args, kws):
        if len(args) != 0:
            return None
        return signature(types.uint32)


def register_vote_sync_intrinsics():
    from numba_cuda_mlir import cuda

    registry.register_global(cuda.all_sync)(AllSyncTemplate)
    registry.register_global(cuda.any_sync)(AnySyncTemplate)
    registry.register_global(cuda.eq_sync)(EqSyncTemplate)
    registry.register_global(cuda.ballot_sync)(BallotSyncTemplate)
    registry.register_global(cuda.match_any_sync)(MatchAnySyncTemplate)
    registry.register_global(cuda.match_all_sync)(MatchAllSyncTemplate)
    registry.register_global(cuda.activemask)(ActivemaskTemplate)


register_vote_sync_intrinsics()


# Cache hint constraint map for PTX inline assembly
CACHE_HINT_CONSTRAINT_MAP = {1: "b", 8: "r", 16: "h", 32: "r", 64: "l", 128: "q"}


def _validate_cache_hint_args(instruction, array, index):
    """Validate arguments for cache hint load/store operations."""
    from numba_cuda_mlir.numba_cuda.core.errors import TypingError

    is_array = isinstance(array, types.Array)
    is_pointer = isinstance(array, types.CPointer)
    if not (is_array or is_pointer):
        raise TypingError(f"{instruction} operates on arrays or pointers. Got type {array}")

    if isinstance(index, types.Integer):
        if is_array and array.ndim != 1:
            raise TypingError(f"Expected {array.ndim} indices, got a scalar")
        return True

    if isinstance(index, types.UniTuple):
        if is_pointer:
            raise TypingError(f"Pointers only support scalar indexing, got tuple of {index.count}")
        if index.count != array.ndim:
            raise TypingError(f"Expected {array.ndim} indices, got {index.count}")
        if isinstance(index.dtype, types.Integer):
            return True

    raise TypingError(f"{index} is not a valid index")


def _validate_cache_hint_dtype(instruction, array):
    """Validate dtype for cache hint operations."""
    from numba_cuda_mlir.numba_cuda.core.errors import TypingError

    dtype = array.dtype
    if not isinstance(dtype, (types.Integer, types.Float)):
        raise TypingError(f"{instruction} requires array of integer or float type, got {dtype}")
    bitwidth = dtype.bitwidth
    if bitwidth not in CACHE_HINT_CONSTRAINT_MAP:
        valid_widths = sorted(CACHE_HINT_CONSTRAINT_MAP.keys())
        raise TypingError(
            f"{instruction} requires array dtype with bitwidth in {valid_widths}, "
            f"got bitwidth {bitwidth}"
        )


class CacheHintLoadTemplate(AbstractTemplate):
    """Base template for cache hint load operations (ldca, ldcg, ldcs, ldlu, ldcv)."""

    instruction = None  # Override in subclasses

    def generic(self, args, kws):
        if len(args) != 2:
            return None
        array, index = args
        _validate_cache_hint_args(self.instruction, array, index)
        _validate_cache_hint_dtype(self.instruction, array)
        return signature(array.dtype, array, index)


class CacheHintStoreTemplate(AbstractTemplate):
    """Base template for cache hint store operations (stcg, stcs, stwb, stwt)."""

    instruction = None  # Override in subclasses

    def generic(self, args, kws):
        if len(args) != 3:
            return None
        array, index, value = args
        _validate_cache_hint_args(self.instruction, array, index)
        _validate_cache_hint_dtype(self.instruction, array)
        return signature(types.void, array, index, value)


class LdcaTemplate(CacheHintLoadTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.ldca
    instruction = "ldca"


class LdcgTemplate(CacheHintLoadTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.ldcg
    instruction = "ldcg"


class LdcsTemplate(CacheHintLoadTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.ldcs
    instruction = "ldcs"


class LdluTemplate(CacheHintLoadTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.ldlu
    instruction = "ldlu"


class LdcvTemplate(CacheHintLoadTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.ldcv
    instruction = "ldcv"


class StcgTemplate(CacheHintStoreTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.stcg
    instruction = "stcg"


class StcsTemplate(CacheHintStoreTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.stcs
    instruction = "stcs"


class StwbTemplate(CacheHintStoreTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.stwb
    instruction = "stwb"


class StwtTemplate(CacheHintStoreTemplate):
    from numba_cuda_mlir import cuda

    key = cuda.stwt
    instruction = "stwt"


def register_cache_hint_intrinsics():
    from numba_cuda_mlir import cuda

    registry.register_global(cuda.ldca)(LdcaTemplate)
    registry.register_global(cuda.ldcg)(LdcgTemplate)
    registry.register_global(cuda.ldcs)(LdcsTemplate)
    registry.register_global(cuda.ldlu)(LdluTemplate)
    registry.register_global(cuda.ldcv)(LdcvTemplate)
    registry.register_global(cuda.stcg)(StcgTemplate)
    registry.register_global(cuda.stcs)(StcsTemplate)
    registry.register_global(cuda.stwb)(StwbTemplate)
    registry.register_global(cuda.stwt)(StwtTemplate)


register_cache_hint_intrinsics()


# Register cuda.bindings.driver.CUtensorMap if available
try:
    from cuda.bindings import driver

    @typeof.typeof_impl.register(driver.CUtensorMap)
    def typeof_cutensormap(val, c):
        """Type inference for cuda.bindings.driver.CUtensorMap objects

        Returns the numba-cuda-mlir CUTensorMap type so these objects can be passed
        to kernels with grid_constant(CUTensorMap) signatures.
        """
        return types.CUTensorMap

except ImportError:
    # cuda-python not available, skip registration
    pass


@registry.register_global(literally)
class LiterallyTemplate(AbstractTemplate):
    """Template for literally() - forces compile-time constant"""

    def generic(self, args, kws):
        if len(args) == 1:
            arg = args[0]
            # The argument must be a literal
            if not isinstance(arg, types.Literal):
                raise ForceLiteralArg(set([0]))
            # Returns the same literal type
            return signature(arg, arg)
        return None


@registry.register_global(literal_unroll)
class LiteralUnrollTemplate(AbstractTemplate):
    """Template for literal_unroll() - forces loop unrolling"""

    def generic(self, args, kws):
        if len(args) == 1:
            arg = args[0]
            # Simply pass through the argument (typically a range or iterable)
            return signature(arg, arg)
        return None


# Also register for numba.cuda.misc.special.literal_unroll (different object)
@registry.register_global(cuda_literal_unroll)
class CudaLiteralUnrollTemplate(AbstractTemplate):
    """Template for cuda literal_unroll() - forces loop unrolling"""

    def generic(self, args, kws):
        if len(args) == 1:
            arg = args[0]
            return signature(arg, arg)
        return None



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.types import Type as NumbaType
from numba_cuda_mlir.numba_cuda import types
import math
from numba_cuda_mlir._mlir import ir

_map_be_type_names_to_fe_types = dict()


class AggregateType(NumbaType):
    """
    Going from a numba type to an MLIR type is easy because llvm stores the types
    for us with llvm.StructType.get_identified. Going from the MLIR type back to the
    numba type is not as easy, so we keep a mapping here.

    Supports bitfields: fields can have (name, type, bit_width) where bit_width specifies
    the number of bits for a bitfield. Bitfields are packed into the smallest possible
    integer type.
    """

    def __init__(
        self,
        name: str,
        fields: list[tuple[str, NumbaType] | tuple[str, NumbaType, int | None]],
    ):
        super().__init__(name)
        # Normalize fields to 3-tuples (name, type, bit_width)
        normalized_fields = [
            (field[0], field[1], None) if len(field) == 2 else field for field in fields
        ]

        self.fields = normalized_fields
        self.__numba_cuda_mlir_pointer_cast__ = True

        # Storage size in bits when this struct is used as a union variant
        # This is set by union() to match C behavior where bitfields in a union
        # struct pack into the union's storage (not the first field's type)
        self.union_storage_bits = None

        # Calculate field layout (contains ALL fields with their bit offsets/widths)
        # This is the single source of truth for field layout information
        self.field_layout = self._calculate_field_layout()

        # This is used by code that checks "if struct.bitfield_info:" to determine
        # if it's a bitfield struct (single storage) vs regular struct (multiple fields)
        self.bitfield_info = {
            name: info for name, info in self.field_layout.items() if info.get("is_bitfield", False)
        }

        AggregateType.record_named_type(name, self)

    @property
    def is_bitfield_struct(self) -> bool:
        """
        Returns True if this struct contains any bitfield fields.

        Bitfield structs are represented in MLIR with a single storage field
        and use shift/mask operations for field access.

        Regular structs have separate MLIR fields for each struct field
        and use extractvalue/insertvalue operations.
        """
        return len(self.bitfield_info) > 0

    def get_bitfield_storage_type(self):
        """
        Determine the storage type for bitfield structs.

        For bitfield structs used as union variants, use the union's storage size.
        Otherwise, default to uint64.

        Returns:
            numba type (uint8/uint16/uint32/uint64)
        """
        if self.union_storage_bits is not None:
            # Use the union's storage size
            if self.union_storage_bits <= 8:
                return types.uint8
            elif self.union_storage_bits <= 16:
                return types.uint16
            elif self.union_storage_bits <= 32:
                return types.uint32
            elif self.union_storage_bits <= 64:
                return types.uint64
            else:
                return types.uint64  # Fallback
        else:
            # Standalone bitfield struct - use uint64
            return types.uint64

    def _calculate_field_layout(self):
        """
        Calculate bit offsets and widths for ALL struct fields (both bitfields and regular).

        Returns a dict: {field_name: {'bit_offset': int, 'bit_width': int, 'underlying_type': type, 'is_bitfield': bool, 'field_index': int}}

        For bitfield structs (fields with explicit bit_width):
        - Fields are packed into a single storage unit at MLIR field index 0
        - bit_offset is relative to the start of the storage
        - field_index is always 0 (the storage field)

        For regular structs (fields without bit_width):
        - Each field is a separate MLIR struct field
        - bit_offset tracks the cumulative bit position (useful for unions)
        - field_index is the position in the MLIR struct (used by insertvalue/extractvalue)

        Supports padding: field_name can be None to indicate anonymous padding bits (bitfields only).

        Current limitation: All bitfields must be at the beginning of the struct and
        must fit within a single 64-bit storage unit.
        """
        field_layout = {}
        bit_offset = 0
        has_seen_regular_field = False
        total_bitfield_bits = 0
        has_bitfields = False
        mlir_field_index = 0  # Track MLIR struct field index for insertvalue/extractvalue

        # For bitfield structs, we always use 64 bits of storage (uint64/i64)
        BITFIELD_STORAGE_BITS = 64

        for field_name, field_type, bit_width in self.fields:
            if bit_width is not None:
                # This is a bitfield (or padding if field_name is None)
                has_bitfields = True

                # Check: No regular fields before this bitfield
                if has_seen_regular_field:
                    field_desc = "padding" if field_name is None else f"field '{field_name}'"
                    raise NotImplementedError(
                        f"TODO: Implement this. Bitfield {field_desc} appears after a regular (non-bitfield) field. "
                        f"Current implementation requires all bitfields to be at the beginning "
                        f"of the struct with no regular fields in between.\n"
                        f"Struct: {self.name}"
                    )

                # Check: Bitfields don't overflow the 64-bit storage unit
                if bit_width <= 0:
                    raise ValueError(
                        f"Bitfield '{field_name}' in struct '{self.name}' has invalid bit_width: {bit_width}. "
                        f"bit_width must be positive."
                    )

                total_bitfield_bits += bit_width
                if total_bitfield_bits > BITFIELD_STORAGE_BITS:
                    field_desc = "padding" if field_name is None else f"field '{field_name}'"
                    raise NotImplementedError(
                        f"TODO: Implement this. Bitfield {field_desc} would overflow the storage unit. "
                        f"Total bitfield size ({total_bitfield_bits} bits) exceeds the "
                        f"storage limit ({BITFIELD_STORAGE_BITS} bits). "
                        f"Current implementation does not support bitfield overflow to "
                        f"multiple storage fields.\n"
                        f"Struct: {self.name}"
                    )

                # Store info only for non-padding fields (field_name is not None)
                # Padding just advances the bit_offset without creating an accessible field
                if field_name is not None:
                    field_layout[field_name] = {
                        "bit_offset": bit_offset,
                        "bit_width": bit_width,
                        "underlying_type": field_type,
                        "is_bitfield": True,
                        "field_index": 0,  # All bitfields are in storage field at index 0
                    }

                bit_offset += bit_width
            else:
                # Regular field (not a bitfield)
                if field_name is None:
                    raise ValueError(
                        f"Regular (non-bitfield) fields cannot have None as field name. "
                        f"None field names are only allowed for bitfield padding.\n"
                        f"Struct: {self.name}"
                    )
                has_seen_regular_field = True

                # Calculate bit width for regular field
                if isinstance(field_type, AggregateType):
                    # Nested struct
                    field_bits = field_type.size_bits
                elif isinstance(field_type, types.UniTuple):
                    # Fixed-size array
                    elem_bits = (
                        field_type.dtype.bitwidth if hasattr(field_type.dtype, "bitwidth") else 64
                    )
                    field_bits = field_type.count * elem_bits
                elif hasattr(field_type, "bitwidth"):
                    field_bits = field_type.bitwidth
                else:
                    # Conservative estimate for unknown types
                    field_bits = 64

                # Track layout for regular fields too
                field_layout[field_name] = {
                    "bit_offset": bit_offset,
                    "bit_width": field_bits,
                    "underlying_type": field_type,
                    "is_bitfield": False,
                    "field_index": mlir_field_index,  # MLIR struct field index for insertvalue/extractvalue
                }

                bit_offset += field_bits
                mlir_field_index += 1  # Each regular field gets its own MLIR struct field

        # For bitfield structs, increment mlir_field_index once for the storage field
        if has_bitfields and mlir_field_index == 0:
            mlir_field_index = 1

        return field_layout

    def _get_type_size_bits(self, numba_type):
        # Map Numba integer types to their bit sizes
        type_size_map = {
            NumbaType.uint8: 8,
            NumbaType.int8: 8,
            NumbaType.uint16: 16,
            NumbaType.int16: 16,
            NumbaType.uint32: 32,
            NumbaType.int32: 32,
            NumbaType.uint64: 64,
            NumbaType.int64: 64,
        }

        if numba_type in type_size_map:
            return type_size_map[numba_type]

        # Fallback: try to get bitwidth attribute if it exists
        if hasattr(numba_type, "bitwidth"):
            return numba_type.bitwidth

        raise ValueError(
            f"Cannot determine bit size for type {numba_type}. "
            f"Bitfields require integer types (uint8, uint16, uint32, uint64, etc.)"
        )

    @property
    def size_bits(self) -> int:
        """
        Calculate the size in bits of this struct.

        For bitfield structs: returns the total bits used, rounded to power of 2 (8, 16, 32, 64).
        For regular structs: sums the sizes of all fields.
        """
        if self.is_bitfield_struct:
            # Bitfield struct: calculate the actual bits used (including padding)
            max_bit_end = sum(bit_width for _, _, bit_width in self.fields if bit_width is not None)
            # Round up to nearest power of 2 (minimum 8)
            return max(8, 2 ** math.ceil(math.log2(max_bit_end)))
        else:
            # Regular struct: sum field sizes (using field_layout)
            total_bits = sum(info["bit_width"] for info in self.field_layout.values())
            return total_bits

    @staticmethod
    def record_named_type(name: str, fe_type: NumbaType):
        """
        Record a named type in the global registry.

        Returns the registered type (either newly registered or existing).
        If a type with the same name exists, returns it (idempotent behavior).
        """
        if name in _map_be_type_names_to_fe_types:
            existing_type = _map_be_type_names_to_fe_types[name]
            # Return existing type for idempotent behavior
            # This allows modules to be imported multiple times without errors
            return existing_type
        _map_be_type_names_to_fe_types[name] = fe_type
        return fe_type

    @staticmethod
    def get_named_type(name: str) -> NumbaType | None:
        return _map_be_type_names_to_fe_types.get(name, None)


class UnionType(NumbaType):
    """
    C-style union type where all variants share the same memory location.

    A union is represented as an integer in MLIR of the size of the largest variant.
    Access to variants is done via bitcasting.

    Variants are stored as (name, type) tuples similar to struct fields.
    """

    def __init__(self, name: str, variants: list[tuple[str, NumbaType]]):
        super().__init__(name)
        self.variants = variants
        self.__numba_cuda_mlir_pointer_cast__ = True
        AggregateType.record_named_type(name, self)

    def get_variant_type(self, variant_name: str) -> NumbaType | None:
        """Get the type of a specific variant by name."""
        for name, vtype in self.variants:
            if name == variant_name:
                return vtype
        return None

    @property
    def size_bits(self) -> int:
        """
        Calculate the storage size in bits for this union (the maximum of all variants).

        This matches C behavior where the union size is the size of its largest member.
        """
        max_bits = 0

        for _variant_name, variant_type in self.variants:
            # Get the size of this variant
            if hasattr(variant_type, "bitwidth"):
                bits = variant_type.bitwidth
            elif isinstance(variant_type, AggregateType):
                # For structs, use their size_bits property
                bits = variant_type.size_bits
            elif isinstance(variant_type, types.UniTuple):
                # UniTuple: count * element_size
                elem_bits = (
                    variant_type.dtype.bitwidth if hasattr(variant_type.dtype, "bitwidth") else 64
                )
                bits = variant_type.count * elem_bits
            else:
                # Default for unknown types
                bits = 64

            max_bits = max(max_bits, bits)

        return max_bits




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class IndexCmpPredicate(IntEnum):
    """index comparison predicate kind"""

    EQ = 0
    NE = 1
    SLT = 2
    SLE = 3
    SGT = 4
    SGE = 5
    ULT = 6
    ULE = 7
    UGT = 8
    UGE = 9

    def __str__(self):
        if self is IndexCmpPredicate.EQ:
            return "eq"
        if self is IndexCmpPredicate.NE:
            return "ne"
        if self is IndexCmpPredicate.SLT:
            return "slt"
        if self is IndexCmpPredicate.SLE:
            return "sle"
        if self is IndexCmpPredicate.SGT:
            return "sgt"
        if self is IndexCmpPredicate.SGE:
            return "sge"
        if self is IndexCmpPredicate.ULT:
            return "ult"
        if self is IndexCmpPredicate.ULE:
            return "ule"
        if self is IndexCmpPredicate.UGT:
            return "ugt"
        if self is IndexCmpPredicate.UGE:
            return "uge"
        raise ValueError("Unknown IndexCmpPredicate enum entry.")



@register_attribute_builder("IndexCmpPredicate", allow_existing=True)
def _indexcmppredicate(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

@register_attribute_builder("index.IndexCmpPredicateAttr")
def _indexcmppredicateattr(x, context):
    return _ods_ir.Attribute.parse(f'#index<cmp_predicate {str(x)}>', context=context)




# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import lru_cache


@lru_cache
def setup_lowering_patches():
    """Apply necessary patches for lowering system."""
    import numba_cuda_mlir.numba_cuda.misc.literal  # noqa: F401


@lru_cache
def register_lowering():
    """Deprecated: Use MLIRTargetContext.load_additional_registries() instead."""
    setup_lowering_patches()



# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .ast.py_type import PyTypeObject


def make_nanobind_metaclass_inheritable():
    from numba_cuda_mlir._mlir import ir

    # Hack to allow us to inherit from nanobind's metaclass type
    # https://github.com/wjakob/nanobind/pull/836
    nb_meta_cls = type(ir.Value)
    _Py_TPFLAGS_BASETYPE = 1 << 10
    PyTypeObject.from_object(nb_meta_cls).tp_flags |= _Py_TPFLAGS_BASETYPE


make_nanobind_metaclass_inheritable()




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class BVCmpPredicate(IntEnum):
    """smt bit-vector comparison predicate"""

    slt = 0
    sle = 1
    sgt = 2
    sge = 3
    ult = 4
    ule = 5
    ugt = 6
    uge = 7

    def __str__(self):
        if self is BVCmpPredicate.slt:
            return "slt"
        if self is BVCmpPredicate.sle:
            return "sle"
        if self is BVCmpPredicate.sgt:
            return "sgt"
        if self is BVCmpPredicate.sge:
            return "sge"
        if self is BVCmpPredicate.ult:
            return "ult"
        if self is BVCmpPredicate.ule:
            return "ule"
        if self is BVCmpPredicate.ugt:
            return "ugt"
        if self is BVCmpPredicate.uge:
            return "uge"
        raise ValueError("Unknown BVCmpPredicate enum entry.")



@register_attribute_builder("BVCmpPredicate", allow_existing=True)
def _bvcmppredicate(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class IntPredicate(IntEnum):
    """smt comparison predicate for integers"""

    lt = 0
    le = 1
    gt = 2
    ge = 3

    def __str__(self):
        if self is IntPredicate.lt:
            return "lt"
        if self is IntPredicate.le:
            return "le"
        if self is IntPredicate.gt:
            return "gt"
        if self is IntPredicate.ge:
            return "ge"
        raise ValueError("Unknown IntPredicate enum entry.")



@register_attribute_builder("IntPredicate", allow_existing=True)
def _intpredicate(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))




# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.dialects import gpu, llvm
from numba_cuda_mlir.logging import trace


def _is_link_target(op: ir.Operation) -> bool:
    return "numba_cuda_mlir.link_target" in op.attributes.keys()


def _ensure_has_regions_and_blocks(op: ir.Operation):
    if len(op.regions) == 0:
        op.regions.append()
    if len(op.regions[0].blocks) == 0:
        op.regions[0].blocks.append()
    return op.regions[-1].blocks[-1]


def _find_link_target(op: ir.Operation) -> ir.Block:
    trace()
    if _is_link_target(op):
        return _ensure_has_regions_and_blocks(op)
    for region in op.regions:
        for block in region.blocks:
            for op in block:
                if target := _find_link_target(op):
                    return target
    return None


def _is_linkable_op(op):
    """Check if an op should be carried across during linking."""
    from numba_cuda_mlir._mlir.dialects import func

    return isinstance(op, (func.FuncOp, gpu.GPUFuncOp, llvm.GlobalOp))


def _extract_linkable_ops(module: ir.Module) -> list[ir.Operation]:
    """Extract all linkable ops (functions, globals) from a module."""
    ops = []
    for op in module.body:
        if _is_linkable_op(op):
            ops.append(op)
        elif isinstance(op, gpu.GPUModuleOp):
            ops.extend(
                nested_op for nested_op in op.regions[0].blocks[0] if _is_linkable_op(nested_op)
            )
    return ops


def _get_op_sym_name(op) -> str | None:
    """Get the symbol name from an op, if it has one."""
    if "sym_name" in op.attributes.keys():
        return op.attributes["sym_name"].value
    return None


def _get_existing_names(block: ir.Block) -> set[str]:
    """Get symbol names of all linkable ops in a block."""
    names = set()
    for op in block:
        if _is_linkable_op(op):
            if name := _get_op_sym_name(op):
                names.add(name)
    return names


def link_inplace(module: ir.Module | str, other: ir.Module | str):
    """
    Pulls all linkable ops (functions and globals) from `other` and inserts
    them into the body of `module`'s gpu.module so they are available.

    TODO: Use junfeng's MLIR-level lto when available.
    """
    trace("linking %s", other)
    if isinstance(module, str):
        module = ir.Module.parse(module)
    if isinstance(other, str):
        other = ir.Module.parse(other)
    link_target = _find_link_target(module.operation)

    existing_names = _get_existing_names(link_target)
    ops = _extract_linkable_ops(other)
    with ir.InsertionPoint(link_target) as ip:
        for op in ops:
            sym_name = _get_op_sym_name(op)
            if sym_name and sym_name in existing_names:
                trace("skipping duplicate symbol %s", sym_name)
                continue
            if sym_name:
                existing_names.add(sym_name)
            cloned = op.clone()
            ip.insert(cloned.detach_from_parent())

    return module
